import aiohttp
from aiohttp import hdrs
from multidict import CIMultiDict
from multidict import CIMultiDictProxy

try:
    import aiosocks
    from aiosocks.connector import SocksConnector
except ImportError:
    pass


class CustomClientResponse(aiohttp.ClientResponse):
    # you have to use this class in response_class parameter of any aiohttp.ClientSession instance
    # example: aiohttp.ClientSession(response_class=CustomClientResponse)
    # read more: https://github.com/Fahreeve/aiovk/issues/3

    async def start(self, connection, read_until_eof=False):
        # vk.com return url like this: http://REDIRECT_URI#access_token=...
        # but aiohttp by default removes all parameters after '#'
        await super().start(connection, read_until_eof)
        headers = CIMultiDict(self.headers)
        location = headers.get(hdrs.LOCATION, None)
        if location is not None:
            headers[hdrs.LOCATION] = location.replace('#', '?')
        self.headers = CIMultiDictProxy(headers)
        self.raw_headers = tuple(headers.items())
        return self


class BaseDriver:
    def __init__(self, timeout=10, loop=None):
        self.timeout = timeout
        self._loop = loop

    async def json(self, url, params, timeout=None):
        '''
        :param params: dict of query params
        :return: dict from json response
        '''
        raise NotImplementedError

    async def get_text(self, url, params, timeout=None):
        '''
        :param params: dict of query params
        :return: http status code, text body of response
        '''
        raise NotImplementedError

    async def get_bin(self, url, params, timeout=None):
        '''
        :param params: dict of query params
        :return: http status code, binary body of response
        '''
        raise NotImplementedError

    async def post_text(self, url, data, timeout=None):
        '''
        :param data: dict pr string
        :return: redirect url and text body of response
        '''
        raise NotImplementedError

    def close(self):
        raise NotImplementedError


class HttpDriver(BaseDriver):
    def __init__(self, timeout=10, loop=None):
        super().__init__(timeout, loop)
        self.session = aiohttp.ClientSession(
            response_class=CustomClientResponse, loop=loop)

    async def json(self, url, params, timeout=None):
        with aiohttp.Timeout(timeout or self.timeout):
            async with self.session.get(url, params=params) as response:
                return await response.json()

    async def get_text(self, url, params, timeout=None):
        with aiohttp.Timeout(timeout or self.timeout):
            response = await self.session.get(url, params=params)
            return response.status, await response.text()

    async def get_bin(self, url, params, timeout=None):
        with aiohttp.Timeout(timeout or self.timeout):
            response = await self.session.get(url, params=params)
            return await response.read()

    async def post_text(self, url, data, timeout=None):
        with aiohttp.Timeout(timeout or self.timeout):
            response = await self.session.post(url, data=data)
            return response.url, await response.text()

    def close(self):
        self.session.close()


class Socks5Driver(HttpDriver):
    connector = SocksConnector

    def __init__(self, address, port, login=None, password=None, timeout=10, loop=None):
        super().__init__(timeout)
        self.close()
        addr = aiosocks.Socks5Addr(address, port)
        if login and password:
            auth = aiosocks.Socks5Auth(login, password=password)
        else:
            auth = None
        conn = self.connector(proxy=addr, proxy_auth=auth, loop=loop)
        self.session = aiohttp.ClientSession(connector=conn, response_class=CustomClientResponse)



import json

from aiovk import API
from aiovk.exceptions import VkLongPollError


class LongPoll:
    def __init__(self, session_or_api, mode, wait=25, version=1, timeout=None):
        if type(session_or_api) == API:
            self.api = session_or_api
        else:
            self.api = API(session_or_api)
        self.timeout = timeout or self.api._session.timeout
        if type(mode) == list:
            mode = sum(mode)
        self.base_params = {
            'version': version,
            'wait': wait,
            'mode': mode,
            'act': 'a_check'
        }
        self.pts = None
        self.ts = None
        self.key = None
        self.base_url = None

    async def _get_long_poll_server(self, need_pts=False):
        response = await self.api('messages.getLongPollServer', need_pts=int(need_pts), timeout=self.timeout)
        self.pts = response.get('pts')
        self.ts = response['ts']
        self.key = response['key']
        self.base_url = 'https://{}'.format(response['server'])

    async def wait(self, need_pts=False):
        if self.base_url is None:
            await self._get_long_poll_server(need_pts)
        params = {
            'ts': self.ts,
            'key': self.key,
        }
        params.update(self.base_params)
        # invalid mymetype from server
        code, response = await self.api._session.driver.get_text(self.base_url, params, timeout=2*self.base_params['wait'])
        if code == 403:
            raise VkLongPollError(403,
                                  'smth weth wrong',
                                  self.base_url + '/',
                                  params
                                  )
        response = json.loads(response)
        failed = response.get('failed')
        if failed is None:
            self.ts = response['ts']
            return response
        if failed == 1:
            self.ts = response['ts']
        elif failed == 4:
            raise VkLongPollError(4,
                                  'An invalid version number was passed in the version parameter',
                                  self.base_url + '/',
                                  params)
        else:
            self.base_url = None
        return await self.wait()

    async def get_pts(self, need_ts=False):
        if self.base_url is None or self.pts is None:
            await self._get_long_poll_server(need_pts=True)
        if need_ts:
            return self.pts, self.ts
        return self.pts



class API:
    def __init__(self, session):
        self._session = session

    def __getattr__(self, method_name):
        return Request(self, method_name)

    async def __call__(self, method_name, **method_kwargs):
        return await getattr(self, method_name)(**method_kwargs)


class Request:
    __slots__ = ('_api', '_method_name', '_method_args')

    def __init__(self, api, method_name):
        self._api = api
        self._method_name = method_name

    def __getattr__(self, method_name):
        return Request(self._api, self._method_name + '.' + method_name)

    async def __call__(self, **method_args):
        timeout = method_args.pop('timeout', None)
        self._method_args = method_args
        return await self._api._session.make_request(self, timeout)



from aiovk.utils import TaskQueue, wait_free_slot


class LimitRateDriverMixin:
    requests_per_period = 3
    period = 1  #seconds

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._queue = TaskQueue(self.requests_per_period, self.period)

    @wait_free_slot
    async def json(self, *args, **kwargs):
        return await super().json(*args, **kwargs)

    @wait_free_slot
    async def get_text(self, *args, **kwargs):
        return await super().get_text(*args, **kwargs)

    @wait_free_slot
    async def get_bin(self, *args, **kwargs):
        return await super().get_bin(*args, **kwargs)

    @wait_free_slot
    async def post_text(self, *args, **kwargs):
        return await super().post_text(*args, **kwargs)

    def close(self):
        super().close()
        self._queue.canel()


class SimpleImplicitSessionMixin:
    """
    Simple implementation of processing captcha and 2factor authorisation
    """

    async def enter_captcha(self, url, sid):
        bytes = await self.driver.get_bin(url, {})
        with open('captcha.jpg', 'wb') as f:
            f.write(bytes)
        return input("Enter captcha: ")

    async def enter_confirmation_сode(self):
        return input('Enter confirmation сode: ')



import urllib.parse
from aiovk.utils import get_request_params


CAPTCHA_IS_NEEDED = 14
AUTHORIZATION_FAILED = 5  # invalid access token


class VkException(Exception):
    pass


class VkAuthError(VkException):
    def __init__(self, error, description, url='', params=''):
        self.error = error
        self.description = description
        self.url = "{}?{}".format(url, urllib.parse.urlencode(params))

    def __str__(self):
        return self.description


class VkCaptchaNeeded(VkException):
    def __init__(self, url, sid):
        self.url = url
        self.sid = sid

    def __str__(self):
        return "You must enter the captcha"


class VkTwoFactorCodeNeeded(VkException):
    def __str__(self):
        return "In order to confirm that you are the owner of this page " \
               "please enter the code provided by the code generating app."


class VkAPIError(VkException):
    def __init__(self, error, url):
        self.error_code = error.get('error_code')
        self.error_msg = error.get('error_msg')
        self.params = get_request_params(error.get('request_params'))
        self.url = url


class VkLongPollError(VkException):
    def __init__(self, error, description, url='', params=''):
        self.error = error
        self.description = description
        self.url = "{}?{}".format(url, urllib.parse.urlencode(params))

    def __str__(self):
        return self.description



import asyncio


class TaskQueue(asyncio.Queue):
    def __init__(self, max_size, period, *args, **kwargs):
        super().__init__(max_size, *args, **kwargs)
        self.period = period

    def _init(self, maxsize):
        super()._init(maxsize)
        for i in range(maxsize):
            self._queue.append(1)
        self.task = asyncio.ensure_future(self.dispatcher(maxsize))

    async def dispatcher(self, maxsize):
        while True:
            await asyncio.sleep(self.period)
            for i in range(maxsize - self.qsize()):
                self.put_nowait(1)

    def canel(self):
        self.task.cancel()


def wait_free_slot(func):
    async def wrapper(self, *args, **kwargs):
        await self._queue.get()
        return await func(self, *args, **kwargs)
    return wrapper


def get_request_params(request_params):
    return {param['key']: param['value'] for param in request_params}



from setuptools import setup, find_packages

with open('README.rst', 'r') as f:
    readme = f.read()

setup(
    name='aiovk',
    version='1.1.0',

    author='Fahreev Eldar',
    author_email='fahreeve@yandex.ru',

    url='https://github.com/Fahreeve/aiovk',
    description='vk.com API python wrapper for asyncio',
    long_description=readme,

    packages=find_packages(),
    install_requires='aiohttp>=2.0.4',

    license='MIT License',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='vk.com api vk wrappper asyncio',
)
